/*This is where we have all our webgl relevant functionality for the plotting routines: */
//{{{ Canvas Initialization
function initCanvas(options) {
	//Initialize open Gl for each canvas, if needed: 
	var canvas=document.getElementById(options.getfieldvalue('canvasid'));
	if (!canvas.initialized) {
		canvas.gl = initWebGL(canvas,options);
		canvas.nodes = [];
		if (canvas.drawHandler)	window.cancelAnimationFrame(canvas.drawHandler);
		draw(canvas,options);
		canvas.initialized = true;
	}
	return canvas;
}
function initWebGL(canvas,options) { //{{{
	var gl;

	try {
		// Try to grab the standard context. If it fails, fallback to experimental.
		gl = canvas.getContext("webgl") || canvas.getContext("experimental-webgl");
	}
	catch(e) {}

	// If we don't have a GL context, give up now
	if (!gl) {
		alert("Unable to initialize WebGL. Your browser may not support it.");
	}
		
	// Enable depth testing
	gl.enable(gl.DEPTH_TEST);
	// Near things obscure far things
	gl.depthFunc(gl.LEQUAL);
	// Enable color blending/overlay
	gl.enable(gl.BLEND);

	// Allocate arrays equal to maximium number of attributes used by any one shader
	gl.enableVertexAttribArray(0);
	gl.enableVertexAttribArray(1);

	// Load shaders and store them in gl object
	gl.shaders = loadShaders(gl);
	
	// Add context state variables
	//TODO:Group variables in objects for organization and naming
	canvas.gl = gl;
	canvas.zoomBounds = options.getfieldvalue('zoomlim',[0.001,100.0]);
	canvas.zoom = clamp(options.getfieldvalue('zoom',1.0), canvas.zoomBounds[0], canvas.zoomBounds[1]);
	canvas.zoomLast = canvas.zoom;
	canvas.cameraMatrix = mat4.create();
	canvas.translation = options.getfieldvalue('origin',[0,0,0.0]);
	canvas.viewPanning = options.getfieldvalue('viewpanning','off') == 'on';
	canvas.view = options.getfieldvalue('view',[0,90]); //0 azimuth - up is north, 90 elevation - looking straight down
	canvas.rotation = canvas.view;
	canvas.rotationAzimuthBounds = options.getfieldvalue('azlim',[0,360]);
	canvas.rotationElevationBounds = options.getfieldvalue('ellim',[-180,180]);
	canvas.controlSensitivity = options.getfieldvalue('controlsensitivity',1);
	canvas.twod = options.getfieldvalue('2d','off') == 'on';
	canvas.moviePlay = true;
	canvas.movieReverse = false;
	canvas.movieIncrement = true;
	canvas.moviefps = options.getfieldvalue('moviefps',5);
	var backgroundcolor = new RGBColor(options.getfieldvalue('backgroundcolor','lightcyan'));
	if (backgroundcolor.ok) canvas.backgroundcolor = [backgroundcolor.r/255.0, backgroundcolor.g/255.0, backgroundcolor.b/255.0, 1.0];
	else throw Error(sprintf("s%s%s\n","initWebGL error message: cound not find out background color for curent canvas ",canvas));
	
	// Add event listeners for canvas
	var displayview = options.getfieldvalue('displayview','off') == 'on';
	var displayzoom = options.getfieldvalue('displayzoom','off') == 'on';
	var mc = new Hammer.Manager(canvas);
	
    mc.add(new Hammer.Pan({threshold:0, pointers:0}));
    mc.add(new Hammer.Pinch({threshold:0})).recognizeWith(mc.get('pan'));
    mc.on("panstart panmove", function (ev) {onPan(ev,canvas,displayview);});
    mc.on("pinchstart pinchmove", function (ev) {onPinch(ev,canvas,displayview);});
	//mc.on("mousewheel DOMMouseScroll", function (ev) {onZoom(ev,canvas,options);});
	
	canvas.addEventListener("mousewheel", function (ev) {onZoom(ev,canvas,displayzoom)}, false);
	canvas.addEventListener("DOMMouseScroll", function (ev) {onZoom(ev,canvas,displayzoom)}, false);
	
	return gl;
} //}}}
function initBuffers(gl,arrays) { //{{{	
	var bufferArray = [];
	for (var i = 0; i < arrays.length; i++) {
		bufferArray[i] = gl.createBuffer();	
		bufferArray[i].itemSize = arrays[i].itemSize;
		bufferArray[i].numItems = arrays[i].length/bufferArray[i].itemSize;
		
		if (bufferArray[i].itemSize > 1) {
			gl.bindBuffer(gl.ARRAY_BUFFER, bufferArray[i]);
			gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(arrays[i]), gl.STATIC_DRAW);
		}
		else {
			//TODO: identify index buffers uniquely (by name)
			gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, bufferArray[i]);
			gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(arrays[i]), gl.STATIC_DRAW);
		}
	}	
	return bufferArray;
} //}}}
function initTexture(gl,imageSource) { //{{{
	var texture = gl.createTexture();
	texture.image = new Image();
	texture.isLoaded = false;
	texture.image.onload = function () {
		handleLoadedTexture(gl,texture);
	}
	texture.image.src = imageSource;
	return texture;
} //}}}
function handleLoadedTexture(gl,texture) { //{{{
	gl.activeTexture(gl.TEXTURE0);
	gl.bindTexture(gl.TEXTURE_2D, texture);
	gl.pixelStorei(gl.UNPACK_FLIP_Y_WEBGL, true);
	gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, texture.image);
	gl.generateMipmap(gl.TEXTURE_2D);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST_MIPMAP_LINEAR);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
	gl.bindTexture(gl.TEXTURE_2D, null);
	texture.isLoaded = true;
} //}}}
function Node(gl,options) { //{{{
	//Returns a Node object that contains default display states for webgl object
	return {buffers:[],
		shader:gl.shaders["colored"]["program"],
		draw:null,
		hideOcean:false,
		level:0,
		useIndexBuffer:true,
		alpha:1.0,
		lineWidth:1.0,
		disableDepthTest:false, 
		enableCullFace:false,
		cullFace:gl.FRONT,
		drawMode:gl.TRIANGLES,
		texture:null,
		translation:vec3.create(),
		rotation:vec3.fromValues(-90, 0, 0),
		scale:vec3.fromValues(1, 1, 1),
		modelMatrix:mat4.create(),
		shaderName:"colored",
		drawOrder:0,
		maskEnabled:false,
		maskHeight:150.0,
		maskColor:vec4.fromValues(0.0, 0.0, 1.0, 1.0),
		enabled:true,
	};
} //}}}
function recalculateModelMatrix(node) { //{{{
	//TODO: move to 0,0,0, rotate,move back to normal space, then apply transform
	var modelMatrix = mat4.create();

	var scaleMatrix = mat4.create();
	mat4.scale(scaleMatrix, scaleMatrix, node["scale"]);
	mat4.multiply(modelMatrix, scaleMatrix, modelMatrix);

	var translationMatrix = mat4.create();
	mat4.translate(translationMatrix, translationMatrix, node["translation"]); //relative translation
	mat4.multiply(modelMatrix, translationMatrix, modelMatrix);
	
	var zRotationMatrix = mat4.create();	
	mat4.rotate(zRotationMatrix, zRotationMatrix, radians(node["rotation"][2]), [0.0, 0.0, 1.0]);
	mat4.multiply(modelMatrix, zRotationMatrix, modelMatrix);
	var yRotationMatrix = mat4.create();	
	mat4.rotate(yRotationMatrix, yRotationMatrix, radians(node["rotation"][1]), [0.0, 1.0, 0.0]);
	mat4.multiply(modelMatrix, yRotationMatrix, modelMatrix);
	var xRotationMatrix = mat4.create();	
	mat4.rotate(xRotationMatrix, xRotationMatrix, radians(node["rotation"][0]), [1.0, 0.0, 0.0]);
	mat4.multiply(modelMatrix, xRotationMatrix, modelMatrix);

	return modelMatrix;
} //}}}
function radians (degrees) { //{{{
  return degrees * Math.PI / 180;
} //}}}
function degrees (radians) { //{{{
  return radians * 180 / Math.PI;
} //}}}
function clamp(value, min, max) { //{{{
	return Math.max(min, Math.min(value, max));
} //}}}
function recover(canvasid,name,defaultvalue) { //{{{
	var canvas  = document.getElementById(canvasid);
	if (canvas && canvas.hasOwnProperty(name)) {
		return canvas[name];
	}
	return defaultvalue;
} //}}}
//}}}
//{{{ Shader Loading
function loadShaders(gl) { //{{{
	var shaderNames = ["colored", "unlit_textured"];
	shaders = {};
	shaders["colored"] = {loaded:false, vsh:{}, fsh:{}};
	shaders["colored"]["vsh"]["string"] = 
		['attribute vec3 aVertexPosition;',
		'attribute vec4 aVertexColor;',
		'',
		'uniform mat4 uMVPMatrix;',
		'uniform float uAlpha;',
		'',
		'varying vec4 vColor;',
		'',
		'void main(void) {',
		'	gl_PointSize = 3.0;',
		'	gl_Position = uMVPMatrix * vec4(aVertexPosition.xyz, 1.0);',
		'	vColor = vec4(aVertexColor.xyz, uAlpha);',
		'}'].join('\n');
	shaders["colored"]["fsh"]["string"] =
		['precision mediump float;',
		'',
		'varying vec4 vColor;',
		'',
		'void main(void) {',
		'	gl_FragColor = vColor;',
		'}'].join('\n');
	shaders["unlit_textured"] = {loaded:false, vsh:{}, fsh:{}};
	shaders["unlit_textured"]["vsh"]["string"] = 
		['attribute vec3 aVertexPosition;',
		'attribute vec2 aTextureCoord;',
		'',
		'uniform mat4 uMVPMatrix;',
		'',
		'varying vec2 vTextureCoord;',
		'varying float vZCoord;',
		'',
		'void main(void) {',
		'	gl_PointSize = 3.0;',
		'	gl_Position = uMVPMatrix * vec4(aVertexPosition.xyz, 1.0);',
		'	vTextureCoord = aTextureCoord;',
		'	vZCoord = aVertexPosition.z;',
		'}'].join('\n');
	shaders["unlit_textured"]["fsh"]["string"] =
		['precision mediump float;',
		'',
		'varying vec2 vTextureCoord;',
		'varying float vZCoord;',
		'',
		'uniform sampler2D uColorSampler;',
		'uniform float uAlpha;',
		'uniform bool uMaskEnabled;',
		'uniform float uMaskHeight;',
		'uniform vec4 uMaskColor;',
		'',
		'void main(void) {',
		'	if (uMaskEnabled && (vZCoord < uMaskHeight)) {',
		'		gl_FragColor = vec4(uMaskColor.xyz, uAlpha);',
		'	}',
		'	else {',
		'		gl_FragColor = vec4(texture2D(uColorSampler, vec2(vTextureCoord.s, vTextureCoord.t)).rgb, uAlpha);',
		'	}',
		'}'].join('\n');
	shaderNames.forEach(function(shaderName){
		shaders[shaderName]["vsh"]["shader"] = getShaderByString(gl, shaders[shaderName]["vsh"]["string"], "vsh");
		shaders[shaderName]["fsh"]["shader"] = getShaderByString(gl, shaders[shaderName]["fsh"]["string"], "fsh");

		shaders[shaderName]["program"] = gl.createProgram();
		gl.attachShader(shaders[shaderName]["program"], shaders[shaderName]["vsh"]["shader"]);
		gl.attachShader(shaders[shaderName]["program"], shaders[shaderName]["fsh"]["shader"]);
		gl.linkProgram(shaders[shaderName]["program"]);

		if (!gl.getProgramParameter(shaders[shaderName]["program"], gl.LINK_STATUS)) {
			alert("Could not initialise shaders");
		}

		var vshStringArray = shaders[shaderName]["vsh"]["string"].split("\n");
		var fshStringArray = shaders[shaderName]["fsh"]["string"].split("\n");
		var line = "";
		var property = "";
		for (var i = 0; i < vshStringArray.length; i++) {
			line = vshStringArray[i];
			if (line.search("attribute") != -1) {
				property = nameFromLine(line);
				shaders[shaderName]["program"][property] = gl.getAttribLocation(shaders[shaderName]["program"], property);
			}
			else if (line.search("uniform") != -1) {
				property = nameFromLine(line);
				shaders[shaderName]["program"][property] = gl.getUniformLocation(shaders[shaderName]["program"], property);
			}
			else if (line.search("void main") != -1) {
				break;
			}
		}
		for (var i = 0; i < fshStringArray.length; i++) {
			line = fshStringArray[i];
			if (line.search("uniform") != -1) {
				property = nameFromLine(line);
				shaders[shaderName]["program"][property] = gl.getUniformLocation(shaders[shaderName]["program"], property);
			}
			else if (line.search("void main") != -1) {
				break;
			}
		}
		shaders[shaderName]["loaded"] = true;
	});
	return shaders;
} //}}}
function getShaderByString(gl,str,type) { //{{{
	var shader;
	if (type == "fsh") {
		shader = gl.createShader(gl.FRAGMENT_SHADER);
	}
	else if (type == "vsh") {
		shader = gl.createShader(gl.VERTEX_SHADER);
	}
	else {
		return null;
	}
	
	gl.shaderSource(shader, str);
	gl.compileShader(shader);

	if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {	
		alert(gl.getShaderInfoLog(shader));
		return null;
	}

	return shader;
} //}}}
function nameFromLine(line) { //{{{
	//returns lowerCamelCase property name from shader line
	var fullName = line.split(" ")[2];
	return fullName.slice(0, fullName.search(";"));
} //}}}
//}}}
//{{{ Interface Functions
function onPan(ev,canvas,displaylog) { //{{{
	ev.preventDefault();
	if (ev.type == 'panstart') {
		canvas.lastDeltaX = 0;
		canvas.lastDeltaY = 0;
	}
	if (ev.srcEvent.shiftKey || ev.pointers.length == 2) {
		if (!canvas.viewPanning) return;
		var deltaX = (canvas.lastDeltaX - ev.deltaX) / canvas.clientWidth / canvas.zoom * -2 * canvas.controlSensitivity;
		var deltaY = (canvas.lastDeltaY - ev.deltaY) / canvas.clientHeight / canvas.zoom * -2 * canvas.controlSensitivity;
		
		if (canvas.twod) {
			canvas.translation[0] += Math.cos(radians(canvas.rotation[0])) * deltaX - Math.sin(radians(0)) * deltaY;
			canvas.translation[1] += Math.sin(radians(canvas.rotation[0])) * deltaX + Math.cos(radians(0)) * deltaY;
		}
		else {
			canvas.translation[0] += Math.cos(radians(canvas.rotation[0])) * deltaX - Math.sin(radians(canvas.rotation[0])) * deltaY;
			canvas.translation[1] += Math.sin(radians(canvas.rotation[0])) * deltaX + Math.cos(radians(canvas.rotation[0])) * deltaY;
		}
	}
	else {
		canvas.rotation[0] += degrees((canvas.lastDeltaX - ev.deltaX) / canvas.clientWidth * -2 * canvas.controlSensitivity);
		canvas.rotation[1] += degrees((canvas.lastDeltaY - ev.deltaY) / canvas.clientHeight * -2 * canvas.controlSensitivity);
		
		if (canvas.rotation[0] > 360) {canvas.rotation[0] -= 360};
		if (canvas.rotation[0] < -360) {canvas.rotation[0] += 360};
		if (canvas.rotation[1] > 180) {canvas.rotation[1] -= 360};
		if (canvas.rotation[1] < -180) {canvas.rotation[1] += 360};
		
		canvas.rotation[0] = clamp(canvas.rotation[0], canvas.rotationAzimuthBounds[0], canvas.rotationAzimuthBounds[1]);
		canvas.rotation[1] = clamp(canvas.rotation[1], canvas.rotationElevationBounds[0], canvas.rotationElevationBounds[1])
	}
	canvas.lastDeltaX = ev.deltaX;
	canvas.lastDeltaY = ev.deltaY;

	if (displaylog) console.log(canvas.rotation);
} //}}}
function onPinch(ev,canvas,displaylog) { //{{{
	ev.preventDefault();
	if (ev.type == 'pinchstart') {
		canvas.zoomLast = canvas.zoom;
	}
	else {
		canvas.zoom = clamp(ev.scale * canvas.zoomLast, canvas.zoomBounds[0], canvas.zoomBounds[1]);
		if (displaylog) console.log(canvas.zoom);
	}
} //}}}
function onZoom(ev,canvas,displaylog) { //{{{
	ev.preventDefault();
	var delta = clamp(clamp(ev.scale || ev.wheelDelta || -ev.detail, -1, 1) * canvas.controlSensitivity * canvas.zoom / 20, -1.0, 1.0);
	canvas.zoom = clamp(canvas.zoom + delta, canvas.zoomBounds[0], canvas.zoomBounds[1]);
	
	if (displaylog) console.log(canvas.zoom);
} //}}}
//}}}
//{{{ Drawing Functions
function updateCameraMatrix(canvas) { //{{{
    //Update view matrix and multiply with projection matrix to get the view-projection (camera) matrix.
	var vMatrix = mat4.create();
	var pMatrix = mat4.create();
	var translateMatrix = mat4.create();
	var rotationMatrix = mat4.create();
	var azimuthRotationMatrix = mat4.create();
	var elevationRotationMatrix = mat4.create();
	var aspectRatio = canvas.clientWidth / canvas.clientHeight;

	if (canvas.twod) {
		mat4.ortho(pMatrix, -aspectRatio/canvas.zoom, aspectRatio/canvas.zoom, -1/canvas.zoom, 1/canvas.zoom, -1.0, 10000.0);
	}
	else {
		mat4.perspective(pMatrix, 60 * Math.PI / 180, aspectRatio, 0.001, 10000.0);
	}
	
	//Apply worldspace translation
	mat4.translate(translateMatrix, translateMatrix, [canvas.translation[0], canvas.translation[2], canvas.translation[1]]);
	mat4.multiply(vMatrix, translateMatrix, vMatrix);
	
	//Calculate rotation around camera focal point about worldspace origin
	if (canvas.twod) {
		mat4.rotate(azimuthRotationMatrix, azimuthRotationMatrix, radians(0), [0, 1, 0]);
		mat4.rotate(elevationRotationMatrix, elevationRotationMatrix, radians(90), [1, 0, 0]);
		mat4.multiply(rotationMatrix, elevationRotationMatrix, azimuthRotationMatrix);
	}
	else {
		mat4.rotate(azimuthRotationMatrix, azimuthRotationMatrix, radians(canvas.rotation[0]), [0, 1, 0]);
		mat4.rotate(elevationRotationMatrix, elevationRotationMatrix, radians(canvas.rotation[1]), [1, 0, 0]);
		mat4.multiply(rotationMatrix, elevationRotationMatrix, azimuthRotationMatrix);
	}
	
	//Apply rotation and scaling transform
	mat4.multiply(vMatrix, rotationMatrix, vMatrix);
	
	//Apply screenspace translation
	mat4.identity(translateMatrix);
	mat4.translate(translateMatrix, translateMatrix, [0.0, 0.0, -1/canvas.zoom]);
	mat4.multiply(vMatrix, translateMatrix, vMatrix);

	//Apply projection matrix to get camera matrix
	mat4.multiply(canvas.cameraMatrix, pMatrix, vMatrix);
}//}}}
function drawSceneGraphNode(canvas,node) { //{{{
	if (!node["enabled"]) {
		return;
	}
	if (node["texture"]) {
		if (!node["texture"]["isLoaded"]) {
			return;
		}
	}
	var gl = canvas.gl;
	bindAttributes(gl, node["shader"], node["buffers"]);
	var mvpMatrix = mat4.create();
	mat4.multiply(mvpMatrix, canvas.cameraMatrix, node["modelMatrix"]);
	gl.uniformMatrix4fv(node["shader"]["uMVPMatrix"], false, mvpMatrix);
	gl.uniform1f(node["shader"]["uAlpha"], node["alpha"]);
	gl.uniform1i(node["shader"]["uMaskEnabled"], node["maskEnabled"]);
	gl.uniform1f(node["shader"]["uMaskHeight"], node["maskHeight"]);
	gl.uniform4fv(node["shader"]["uMaskColor"], node["maskColor"]);
	if (node["texture"]) {
		gl.activeTexture(gl.TEXTURE0);
		gl.bindTexture(gl.TEXTURE_2D, node["texture"]);
		gl.uniform1i(node["shader"]["uColorSampler"], 0);	
	}
	if (node["disableDepthTest"]) {
		gl.disable(gl.DEPTH_TEST);
	}
	gl.lineWidth(node["lineWidth"]);
	gl.blendFunc (gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);
	if  (node["useIndexBuffer"] == true) {
		gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, node["buffers"][node["buffers"].length - 1]);
		gl.drawElements(node["drawMode"], node["buffers"][node["buffers"].length - 1].numItems, gl.UNSIGNED_SHORT, 0);
	}
	else {
		gl.drawArrays(node["drawMode"], 0, node["buffers"][0].numItems);
	}	
	gl.enable(gl.DEPTH_TEST);
} //}}}
function bindAttributes(gl,shaderProgram,bufferArray) { //{{{
	gl.useProgram(shaderProgram);
	var arrayNumber = 0;
	for (var propertyName in shaderProgram) {
		if (propertyName[0] == "a") {
			if (bufferArray[arrayNumber].itemSize > 1) {
				gl.bindBuffer(gl.ARRAY_BUFFER, bufferArray[arrayNumber]);
				gl.vertexAttribPointer(shaderProgram[propertyName], bufferArray[arrayNumber].itemSize, gl.FLOAT, false, 0, 0);
				arrayNumber++;
			}
		}
	}
} //}}}
function draw(canvas,options) { //{{{
	if (canvas.nodes.length < 1) {
		canvas.drawHandler = window.requestAnimationFrame(function(time) {draw(canvas,options)});
		return;
	}
	// Ensure canvas and gl viewport sizes are the same
	var displayWidth  = canvas.clientWidth;
	var displayHeight = canvas.clientHeight;
	if (canvas.width  != displayWidth || canvas.height != displayHeight) {
		canvas.width  = displayWidth;
		canvas.height = displayHeight;
		canvas.gl.viewport(0, 0, canvas.width, canvas.height);
	}
	
	if (canvas.textcanvas) canvas.textcanvas.draw(canvas);

	var gl = canvas.gl;
	gl.clearColor(canvas.backgroundcolor[0], canvas.backgroundcolor[1], canvas.backgroundcolor[2], canvas.backgroundcolor[3]);
		
	// Skip drawing of new frame if any texture is not yet loaded
	var nodes = canvas.nodes;
	for (var node in nodes) {
		if (nodes[node]["texture"] && !nodes[node]["texture"]["isLoaded"]) {
			canvas.drawHandler = window.requestAnimationFrame(function(time) {draw(canvas,options)});
			return;
		}
	}
	// Else, clear the color as well as the depth buffer for new frame
	gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
	
	updateCameraMatrix(canvas);
	
	var drawPassNumber = 2;
	for (var i = drawPassNumber - 1; i >= 0; i--) {
		for (var node in nodes) {
			if (nodes[node]["drawOrder"] == i) drawSceneGraphNode(canvas,nodes[node]);
		}
	}
	canvas.drawHandler = window.requestAnimationFrame(function(time) {draw(canvas,options)});
} //}}}
//}}}
